Rem * Example 012 TERRAIN Rendering

This tutorial will briefly show how to use the terrain renderer of Irrlicht. It
will also show the terrain renderer triangle selector to be able to do
collision detection with terrain.

Note that the Terrain Renderer in Irrlicht is based on Spintz'
GeoMipMapSceneNode, lots of thanks go to him. DeusXL provided a new elegant
simple solution for building larger area on small heightmaps . terrain
smoothing.

In the beginning there is nothing special. We include the needed header files
and create an event listener to listen if the user presses a key: The 'W' key
switches to wireframe mode, the 'P' key to pointcloud mode, and the 'D' key
toggles between solid and detail mapped material.
EndRem
SuperStrict
Framework Irrlicht.Core

Type MyEventReceiver Extends IEventReceiver

	Method Init:MyEventReceiver(terrain:ISceneNode, skybox:ISceneNode, skydome:ISceneNode)
		Self.Terrain = terrain
		Self.Skybox = skybox
		Self.Skydome = skydome
		Self.skybox.setVisible(True)
		Self.skydome.setVisible(False)
		Return Self
	EndMethod

	Method OnEvent:Int(event:SEvent)
	
		' check If user presses the key 'W' or 'D'
		If event.getEventType()=EET_KEY_INPUT_EVENT And Not event.getKeyPressedDown()
		
			Local key:Int=event.getKeyInputKey()
			
			Select key
				' switch wire frame mode
				Case EKEY_KEY_W  
					Terrain.setMaterialFlag(EMF_WIREFRAME, ..
						Not Terrain.getMaterial(0).getWireframe())
					Terrain.setMaterialFlag(EMF_POINTCLOUD, False)
					Return True
				Case EKEY_KEY_P ' switch wire frame mode
					Terrain.setMaterialFlag(EMF_POINTCLOUD, ..
						Not Terrain.getMaterial(0).getPointCloud())
					Terrain.setMaterialFlag(EMF_WIREFRAME, False)
					Return True					
				Case EKEY_KEY_D ' toggle detail map	
					If Terrain.getMaterial(0).getMaterialType() = EMT_SOLID
						Terrain.setMaterialType(EMT_DETAIL_MAP)
					Else
						Terrain.setMaterialType(EMT_SOLID)
					EndIf
					Return True				
				Case EKEY_KEY_S ' toggle skies
					showBox = Not showBox
					Skybox.setVisible(showBox)
					Skydome.setVisible(Not showBox)
					Return True
			EndSelect			
		EndIf
		
		Return False
	EndMethod

	Field Terrain:ISceneNode
	Field Skybox:ISceneNode
	Field Skydome:ISceneNode
	Field showBox:Int

	' we must override the generate function in order for instantiation to work properly
	Function generate:IEventReceiver()
		Return New MyEventReceiver
	EndFunction
EndType

Rem
The start of the main function starts like in most other example. We ask the user
for the desired renderer and start it up.  This time with the advanced parameter handling.
EndRem

' let user Select driver Type
Local driverType:Int=EDT_NULL

Local i:String=Input("Please select the driver you want for this example:"+Chr(13)+Chr(10)+ ..
	" (a) Direct3D 9.0c"+Chr(13)+Chr(10)+" (b) Direct3D 8.1"+Chr(13)+Chr(10)+ ..
	" (c) OpenGL 1.5"+Chr(13)+Chr(10)+" (d) Software Renderer"+Chr(13)+Chr(10)+ ..
	" (e) Burning's Software Renderer"+Chr(13)+Chr(10)+" (f) NullDevice"+Chr(13)+Chr(10)+ ..
	" (otherKey) Exit"+Chr(13)+Chr(10)+Chr(13)+Chr(10))

Select i
	Case "a" driverType=EDT_DIRECT3D9
	Case "b" driverType=EDT_DIRECT3D8
	Case "c" driverType=EDT_OPENGL
	Case "d" driverType=EDT_SOFTWARE
	Case "e" driverType=EDT_BURNINGVIDEO
	Case "f" driverType=EDT_NULL
	Default Return 
EndSelect

' Create device with full flexibility over creation parameters
' you can add more parameters If desired, check irr::SIrrlichtCreationParameters

Local params:SIrrlichtCreationParameters = SIrrlichtCreationParameters.Create()
params.setDriverType(driverType)
params.setWindowSize(_DIMENSION2DI(640, 480))

Local device:IrrlichtDevice = ..
	IrrlichtDevice.createFromParams(params)

If Not device Return ' could Not create selected driver.
	
Rem
First, we add standard stuff to the scene: A nice irrlicht engine
logo, a small help text, a user controlled camera, and we disable
the mouse cursor.
EndRem

Local driver:IVideoDriver=device.getVideoDriver()
Local smgr:ISceneManager=device.getSceneManager()
Local env:IGUIEnvironment=device.getGUIEnvironment()

driver.setTextureCreationFlag(ETCF_ALWAYS_32_BIT, True)

' add irrlicht logo
env.addImageFromTexture(driver.getTexture("../media/irrlichtlogo2.png"), ..
	_POSITION2DI(10,10))
	
'set other font
env.getSkin().setFont(env.getFont("../media/fontlucida.png"))

' add some help text
Local text:IGUIStaticText = env.addStaticText( ..
	"Press 'W' to change wireframe mode" + Chr(10) + "Press 'D' to toggle detail map" + Chr(10) + "Press 'S' to toggle skybox/skydome", ..
	_RECTI(10,421,250,475), True, True, Null, -1, True)

' add camera
Local camera:ICameraSceneNode = ..
	smgr.addCameraSceneNodeFPS(Null,100.0,1.2)

camera.setPosition(_VECTOR3DF(2700*2,255*2,2600*2))
camera.setTarget(_VECTOR3DF(2397*2,343*2,2700*2))
camera.setFarValue(42000.0)

' disable mouse cursor
device.getCursorControl().setVisible(False)

Rem
Here comes the terrain renderer scene node: We add it just like any
other scene node to the scene using
ISceneManager::addTerrainSceneNode(). The only parameter we use is a
file name to the heightmap we use. A heightmap is simply a gray scale
texture. The terrain renderer loads it and creates the 3D terrain from
it.

To make the terrain look more big, we change the scale factor of
it to (40, 4.4, 40). Because we don't have any dynamic lights in the
scene, we switch off the lighting, and we set the file
terrain-texture.jpg as texture for the terrain and detailmap3.jpg as
second texture, called detail map. At last, we set the scale values for
the texture: The first texture will be repeated only one time over the
whole terrain, and the second one (detail map) 20 times.
EndRem

' add terrain scene node
Local terrain:ITerrainSceneNode = smgr.addTerrainSceneNodeFromFile( .. 
	"../media/terrain-heightmap.bmp", .. 
	Null, ..									'  parent node
	-1, ..										'  node id
	_VECTOR3DF(0.0, 0.0, 0.0), ..				'  position
	_VECTOR3DF(0.0, 0.0, 0.0), ..				'  rotation
	_VECTOR3DF(40.0, 4.4, 40.0), ..				'  scale
	_SCOLOR( 255, 255, 255, 255 ), ..			'  vertexColor,
	5, ..										'  maxLOD
	ETPS_17, ..									'  patchSize
	4 ..										'  smoothFactor
	)

terrain.setMaterialFlag(EMF_LIGHTING, False)

terrain.setMaterialTexture(0, ..
	driver.getTexture("../media/terrain-texture.jpg"))
terrain.setMaterialTexture(1, ..
	driver.getTexture("../media/detailmap3.jpg"))

terrain.setMaterialType(EMT_DETAIL_MAP)

terrain.ScaleTexture(1.0, 20.0)

Rem
To be able to do collision with the terrain, we create a triangle selector.
If you want to know what triangle selectors do, just take a look into the
collision tutorial. The terrain triangle selector works together with the
terrain. To demonstrate this, we create a collision response animator
and attach it to the camera, so that the camera will not be able to fly
through the terrain.
EndRem

' Create triangle selector For the terrain	
Local selector:ITriangleSelector ..
	= smgr.createTerrainTriangleSelector(terrain, 0)
terrain.setTriangleSelector(selector)

'  create collision response animator And attach it To the camera
Local anim:ISceneNodeAnimator = smgr.createCollisionResponseAnimator( ..
	selector, camera, _VECTOR3DF(60,100,60), ..
	_VECTOR3DF(0,0,0), ..
	_VECTOR3DF(0,50,0))
selector.drop()
camera.addAnimator(anim)
anim.drop()

Rem 
If you need access to the terrain data you can also do this directly via the following code fragment.
EndRem
Local buffer:CDynamicMeshBuffer = CDynamicMeshBuffer.Create(EVT_2TCOORDS, EIT_16BIT)
terrain.getMeshBufferForLOD(buffer, 0)
Local data:S3DVertex2TCoords[] = S3DVertex2TCoords[](buffer.getVertexBuffer().pointer())
' Work on data Or get the IndexBuffer with a similar call.
buffer.drop() ' When done drop the buffer again.

Rem
To make the user be able to switch between normal and wireframe mode,
we create an instance of the event reciever from above and let Irrlicht
know about it. In addition, we add the skybox which we already used in
lots of Irrlicht examples and a skydome, which is shown mutually
exclusive with the skybox by pressing 'S'.
EndRem

' Create skybox And skydome
driver.setTextureCreationFlag(ETCF_CREATE_MIP_MAPS, False)

Local skybox:ISceneNode = smgr.addSkyBoxSceneNode( ..
	driver.getTexture("../media/irrlicht2_up.jpg"), ..
	driver.getTexture("../media/irrlicht2_dn.jpg"), ..
	driver.getTexture("../media/irrlicht2_lf.jpg"), ..
	driver.getTexture("../media/irrlicht2_rt.jpg"), ..
	driver.getTexture("../media/irrlicht2_ft.jpg"), ..
	driver.getTexture("../media/irrlicht2_bk.jpg"))
Local skydome:ISceneNode=smgr.addSkyDomeSceneNode(driver.getTexture("../media/skydome.jpg"),16,8,0.95,2.0)

driver.setTextureCreationFlag(ETCF_CREATE_MIP_MAPS, True)

' Create event receiver
Local receiver:MyEventReceiver = ..
	MyEventReceiver(MyEventReceiver.Create(MyEventReceiver.generate)).Init(terrain, skybox, skydome)
device.setEventReceiver(receiver)

Rem
That's it, draw everything.
EndRem

Local lastFPS:Int = -1

While(device.run())
	If (device.isWindowActive())
		driver.beginScene(True, True, Null )
	
		smgr.drawAll()
		env.drawAll()
	
		driver.endScene()
	
		'  display frames per second in window title
		Local fps:Int = driver.getFPS()
		If (lastFPS <> fps)
			Local str:String = "Terrain Renderer - Irrlicht Engine ["
			str :+ driver.getName()
			str :+ "] FPS:"
			str :+ fps
			'  Also Print terrain height of current camera position
			'  We can use camera position because terrain is located at coordinate origin
			str :+ " Height: "
			str :+ terrain.getHeight(camera.getAbsolutePosition().getX(), ..
				camera.getAbsolutePosition().getZ())
	
			device.setWindowCaption(str)
			lastFPS = fps
		EndIf
	EndIf
Wend

device.drop()

Rem
Now you know how to use terrain in Irrlicht.
EndRem
